/*
 * SPDX-License-Identifier: Apache-2.0
 * Copyright (C) 2022, Intel Corporation
 * Description:
 * Example to use LCD Display in Cyclone V SoC FPGA devkit
 */

#include <zephyr/sys/printk.h>
#include <stdio.h>
#include <string.h>
#include <zephyr/drivers/i2c.h>
#include <zephyr/kernel.h>
#include "commands.h"

#include <zephyr/device.h>
#include <zephyr/drivers/gpio.h>

#define I2C_INST DT_NODELABEL(arduino_i2c)

// Define the RTC register addresses based on RTC module documentation
#define RTC_ADDRESS            0x32  // Replace with your RTC module's I2C address
#define RTC_SEC_REGISTER       0x10
#define RTC_MIN_REGISTER       0x11
#define RTC_HOUR_REGISTER      0x12
#define RTC_DAY_REGISTER       0x14
#define RTC_MONTH_REGISTER     0x15
#define RTC_YEAR_REGISTER      0x16
#define RTC_MIN_ALARM_REGISTER 0x18
#define RTC_HOUR_ALARM_REGISTER 0x19
#define RTC_DAY_WEEK_ALARM_REGISTER 0x1A
#define RTC_WADA_REGISTER      0x1D
#define RTC_STATUS_REGISTER    0x1E
#define RTC_AIE_REGISTER       0x1F

#define RTC_ALARM_FLAG_BIT     3     // Example bit for the Alarm Flag (AF) in the status register
#define RTC_ALARM_ENABLE_BIT   7     // Example bit for Alarm Enable (AE) in alarm registers

// #define RTC_ALARM DT_ALIAS(intrtc)

// static const struct gpio_dt_spec rtc_pin = GPIO_DT_SPEC_GET_OR(RTC_ALARM, gpios, {0});
// static struct gpio_callback rtc_gpio_cb;

#define RTC_INT_PIN 1  // Pin number for P1.01
#define GPIO_LABEL "gpio@842800"  // Label from device tree

#define MY_STACK_SIZE 1024  // Define an appropriate size for the stack

K_THREAD_STACK_DEFINE(my_stack_area, MY_STACK_SIZE);  // Allocate the stack area

static const struct device *gpio_dev;
static struct gpio_callback rtc_gpio_cb;

// static K_SEM_DEFINE(RTC_sem, 0, 1);
static struct k_work alarm_work;
static struct k_work_q my_work_q;

static const struct device *i2c = DEVICE_DT_GET(I2C_INST);

typedef struct {
    uint8_t seconds;  // BCD format
    uint8_t minutes;  // BCD format
    uint8_t hours;    // BCD format
    uint8_t day;      // BCD format
    uint8_t month;    // BCD format
    uint8_t year;     // BCD format (year 00-99 corresponds to 2000-2099)
} rtc_time_t;

void rtc_get_alarm(rtc_time_t *alarm_time);
bool occured = false;

void rtc_gpio_callback(const struct device *dev, struct gpio_callback *cb, uint32_t pins)
{
    printk("RTC interrupt triggered on P1.01\n");
    // occured = true;
   	// k_sem_give(&RTC_sem);

//  // Disable the interrupt temporarily
//     gpio_pin_interrupt_configure_dt(&gpio_dev, GPIO_INT_DISABLE);

//     // Schedule the work item to clear the alarm in a safe context

    // k_work_submit_to_queue(&my_work_q, &alarm_work);
    clear_rtc_alarm_interrupt();
}

/* Work item handler for clearing the RTC alarm */
void alarm_work_handler(struct k_work *work)
{
    clear_rtc_alarm_interrupt();  // Perform I2C operations outside of the interrupt context
}

void configure_rtc_gpio_interrupt(void)
{
    int ret;

    /* Get the GPIO device */
    gpio_dev = device_get_binding(GPIO_LABEL);
    if (!gpio_dev) {
        printk("Error: Failed to get GPIO device\n");
        return;
    }

    /* Configure the GPIO pin as input */
    ret = gpio_pin_configure(gpio_dev, RTC_INT_PIN, GPIO_INPUT | GPIO_PULL_UP);
    if (ret != 0) {
        printk("Error %d: Failed to configure GPIO pin\n", ret);
        return;
    }

    /* Configure the GPIO pin interrupt */
    ret = gpio_pin_interrupt_configure(gpio_dev, RTC_INT_PIN, GPIO_INT_LEVEL_LOW);
    if (ret != 0) {
        printk("Error %d: Failed to configure GPIO interrupt\n", ret);
        return;
    }

    /* Initialize the GPIO callback */
    gpio_init_callback(&rtc_gpio_cb, rtc_gpio_callback, BIT(RTC_INT_PIN));
    gpio_add_callback(gpio_dev, &rtc_gpio_cb);

    printk("RTC interrupt configured on P1.01\n");
}

void rtc_write_register(uint8_t reg, uint8_t value) {
    uint8_t buf[2] = {reg, value};
    int ret = i2c_write(i2c, buf, sizeof(buf), RTC_ADDRESS);

    if (ret == 0) {
        // printk("Register 0x%02X write successfully: 0x%02X\n", reg, value);
    } else {
        printk("Failed to write register 0x%02X\n", reg);
    }
}

int rtc_read_register(uint8_t reg, uint8_t *value) {
    int ret = i2c_write(i2c, &reg, 1, RTC_ADDRESS);

    if (ret != 0) {
        printk("Failed to set register 0x%02X for reading\n", reg);
        return ret;
    }

    return i2c_read(i2c, value, 1, RTC_ADDRESS);
}

void rtc_set_time_date(rtc_time_t *time) 
{
    rtc_write_register(RTC_SEC_REGISTER, ((time->seconds / 10) << 4) | (time->seconds % 10));
    rtc_write_register(RTC_MIN_REGISTER, ((time->minutes / 10) << 4) | (time->minutes % 10));
    rtc_write_register(RTC_HOUR_REGISTER, ((time->hours / 10) << 4) | (time->hours % 10));
    rtc_write_register(RTC_DAY_REGISTER, ((time->day / 10) << 4) | (time->day % 10));
    rtc_write_register(RTC_MONTH_REGISTER, ((time->month / 10) << 4) | (time->month % 10));
    rtc_write_register(RTC_YEAR_REGISTER, ((time->year / 10) << 4) | (time->year % 10));
}

int rtc_get_time_date(rtc_time_t *time) 
{
    uint8_t reg_val;
    int ret;

    ret = rtc_read_register(RTC_SEC_REGISTER, &reg_val);
    if (ret) return ret;
    time->seconds = ((reg_val >> 4) * 10) + (reg_val & 0x0F);

    ret = rtc_read_register(RTC_MIN_REGISTER, &reg_val);
    if (ret) return ret;
    time->minutes = ((reg_val >> 4) * 10) + (reg_val & 0x0F);

    ret = rtc_read_register(RTC_HOUR_REGISTER, &reg_val);
    if (ret) return ret;
    time->hours = ((reg_val >> 4) * 10) + (reg_val & 0x0F);

    ret = rtc_read_register(RTC_DAY_REGISTER, &reg_val);
    if (ret) return ret;
    time->day = ((reg_val >> 4) * 10) + (reg_val & 0x0F);

    ret = rtc_read_register(RTC_MONTH_REGISTER, &reg_val);
    if (ret) return ret;
    time->month = ((reg_val >> 4) * 10) + (reg_val & 0x0F);

    ret = rtc_read_register(RTC_YEAR_REGISTER, &reg_val);
    if (ret) return ret;
    time->year = ((reg_val >> 4) * 10) + (reg_val & 0x0F);

    return 0;
}

bool rtc_is_alarm_triggered(void) 
{
    uint8_t reg_1E;
    if (rtc_read_register(RTC_STATUS_REGISTER, &reg_1E) != 0) {
        printk("Failed to read STATUS register\n");
        return false;
    }

    bool alarm_triggered = (reg_1E & (1 << RTC_ALARM_FLAG_BIT)) != 0;
    printk("Alarm Status: %x %s\n", reg_1E, alarm_triggered ? "Triggered" : "Not Triggered");

    return alarm_triggered;
}

void check_alarm_status(void) 
{
    rtc_time_t alarm_time;

    // Check if the alarm has been triggered
    if (rtc_is_alarm_triggered()) {
        printk("Alarm has been triggered!\n");
            clear_rtc_alarm_interrupt();

    } else {
        // printk("Alarm has not been triggered.\n");
    }
}

void clear_rtc_alarm_interrupt(void) {
    uint8_t reg_status;

    // Read the STATUS register to clear the alarm flag
    if (rtc_read_register(RTC_STATUS_REGISTER, &reg_status) != 0) {
        printk("Failed to read STATUS register\n");
        // return;
    }

    reg_status &= ~(1 << RTC_ALARM_FLAG_BIT); // Clear the Alarm Flag bit
    rtc_write_register(RTC_STATUS_REGISTER, reg_status);
}

void rtc_set_alarm(const rtc_time_t *alarm_time, bool enable_day_alarm, bool enable_week_alarm) 
{
    if (alarm_time == NULL) {
        printk("Invalid alarm time structure\n");
        return;
    }

    uint8_t min_alarm = ((alarm_time->minutes / 10) << 4) | (alarm_time->minutes % 10);
    rtc_write_register(RTC_MIN_ALARM_REGISTER, min_alarm);
    // printk("Register 0x18 write successfully: 0x%02X\n", min_alarm);

    uint8_t hour_alarm = ((alarm_time->hours / 10) << 4) | (alarm_time->hours % 10);
    rtc_write_register(RTC_HOUR_ALARM_REGISTER, hour_alarm);
    // printk("Register 0x19 write successfully: 0x%02X\n", hour_alarm);

    uint8_t day_week_alarm = 0;
    if (enable_day_alarm) {
        day_week_alarm = ((alarm_time->day / 10) << 4) | (alarm_time->day % 10);
    } else if (enable_week_alarm) {
        day_week_alarm = alarm_time->month & 0x07; // Week value should be between 0-7
    }
    rtc_write_register(RTC_DAY_WEEK_ALARM_REGISTER, day_week_alarm);
    // printk("Register 0x1A write successfully: 0x%02X\n", day_week_alarm);

	/*Write wada in ext register*/
    uint8_t reg_1D;
    if (rtc_read_register(RTC_WADA_REGISTER, &reg_1D) != 0) {
        printk("Failed to read WADA register\n");
        return;
    }
	// printk("Ext register %x\n", reg_0A);
    if (enable_day_alarm) 
	{
        reg_1D |= (1 << 3); // Clear WADA bit to select DAY Alarm
    } 
	// else if (enable_week_alarm) {
    //     reg_0A |= (1 << 0);  // Set WADA bit to select WEEK Alarm
    // }
    rtc_write_register(RTC_WADA_REGISTER, reg_1D);
    // printk("Register 0x0A write successfully: 0x%02X\n", reg_0A);

    uint8_t min_alarm_value;
    if (rtc_read_register(RTC_MIN_ALARM_REGISTER, &min_alarm_value) != 0) {
        printk("Failed to read MIN Alarm register\n");
        return;
    }
    min_alarm_value &= ~(1 << RTC_ALARM_ENABLE_BIT);
	// printk("min_alarm_value %x\n", min_alarm_value);
    rtc_write_register(RTC_MIN_ALARM_REGISTER, min_alarm_value);

    uint8_t hour_alarm_value;
    if (rtc_read_register(RTC_HOUR_ALARM_REGISTER, &hour_alarm_value) != 0) {
        printk("Failed to read HOUR Alarm register\n");
        return;
    }
    hour_alarm_value &= ~(1 << RTC_ALARM_ENABLE_BIT);
	// printk("hour_alarm_value %x\n", hour_alarm_value);
    rtc_write_register(RTC_HOUR_ALARM_REGISTER, hour_alarm_value);

    uint8_t day_week_alarm_value;
    if (rtc_read_register(RTC_DAY_WEEK_ALARM_REGISTER, &day_week_alarm_value) != 0) {
        printk("Failed to read DAY/WEEK Alarm register\n");
        return;
    }
    day_week_alarm_value &= ~(1 << RTC_ALARM_ENABLE_BIT);
	// printk("day_week_alarm_value %x\n", day_week_alarm_value);
    rtc_write_register(RTC_DAY_WEEK_ALARM_REGISTER, day_week_alarm_value);

    uint8_t reg_1F;
    if (rtc_read_register(RTC_AIE_REGISTER, &reg_1F) != 0) {
        printk("Failed to read AIE register\n");
        return;
    }
    reg_1F |= (1 << 3); // Set AIE bit
	// printk("reg_1F %x\n", reg_1F);
    rtc_write_register(RTC_AIE_REGISTER, reg_1F);
    // printk("Register 0x1F write successfully: 0x%02X\n", reg_1F);

    uint8_t reg_1E;
    if (rtc_read_register(RTC_STATUS_REGISTER, &reg_1E) != 0) {
        printk("Failed to read STATUS register\n");
        return;
    }
    reg_1E &= ~(1 << RTC_ALARM_FLAG_BIT); // Clear the Alarm Flag
    rtc_write_register(RTC_STATUS_REGISTER, reg_1E);
    printk("Register 0x1E write successfully: 0x%x\n", reg_1E);
}

void rtc_get_alarm(rtc_time_t *alarm_time) 
{
    if (alarm_time == NULL) {
        printk("Invalid alarm time structure\n");
        return;
    }

    uint8_t min_alarm;
    if (rtc_read_register(RTC_MIN_ALARM_REGISTER, &min_alarm) != 0) {
        printk("Failed to read MIN Alarm register\n");
        return;
    }
    alarm_time->minutes = ((min_alarm >> 4) * 10) + (min_alarm & 0x0F);

    uint8_t hour_alarm;
    if (rtc_read_register(RTC_HOUR_ALARM_REGISTER, &hour_alarm) != 0) {
        printk("Failed to read HOUR Alarm register\n");
        return;
    }
    alarm_time->hours = ((hour_alarm >> 4) * 10) + (hour_alarm & 0x0F);

    uint8_t day_week_alarm;
    if (rtc_read_register(RTC_DAY_WEEK_ALARM_REGISTER, &day_week_alarm) != 0) {
        printk("Failed to read DAY/WEEK Alarm register\n");
        return;
    }
    if (true) { // Assuming you have a function to determine the alarm type
        alarm_time->day = ((day_week_alarm >> 4) * 10) + (day_week_alarm & 0x0F);
    } else {
        alarm_time->month = day_week_alarm & 0x07; // Week value
    }

    printk("Alarm Time: %02d:%02d:%02d\n", alarm_time->hours, alarm_time->minutes, alarm_time->seconds);
    printk("Alarm Date: %02d/%02d/%02d\n", alarm_time->day, alarm_time->month, alarm_time->year);
}

void print_time_date(rtc_time_t *time) {
    printk(": %02d:%02d:%02d ", time->hours, time->minutes, time->seconds);
    printk(": %02d/%02d/20%02d\n", time->day, time->month, time->year);
}

int main(void)
{
	printk("Hello World! %s\n", CONFIG_BOARD);
	static const unsigned char string[] = "Hello world!";
	int len = strlen((const unsigned char *) &string);

	if (!device_is_ready(i2c)) {
		printf("i2c is not ready!\n");
	} else {
		printf("i2c is ready\n");
	}

    configure_rtc_gpio_interrupt();

    k_work_queue_start(&my_work_q, my_stack_area, K_THREAD_STACK_SIZEOF(my_stack_area), -1, NULL);

    k_work_init(&alarm_work, alarm_work_handler);
   // Define the current time and alarm time
    rtc_time_t current_time = {
        .hours = 12, .minutes = 34, .seconds = 50,
        .day = 5, .month = 9, .year = 24  // 12:34:56, 5th Sep 2024
    };

    rtc_time_t alarm_time = {
        .hours = 12, .minutes = 36, .seconds = 00,  // Set alarm for 12:35:00
        .day = 5, .month = 9, .year = 24
    };

	rtc_time_t getcurrent_time[] = {};
	rtc_time_t getalarm_time[] = {};
    // Initialize the RTC
    printf("Initializing RTC...\n");

    printf("Set Time...\n");

    // Set the current date and time
    rtc_set_time_date(&current_time);
	print_time_date(&current_time);

    printf("Get Time...\n");

   	rtc_get_time_date(&getcurrent_time);
	print_time_date(&getcurrent_time);

    printf("set alarm Time...\n");

    // Set the alarm
    rtc_set_alarm(&alarm_time, true, false);
	print_time_date(&alarm_time);

    // printf("get alarm Time...\n");

	// rtc_get_alarm(&getalarm_time);
	// print_time_date(&getalarm_time);

    printf("Timer configured and enabled.\n");

    // Main loop
    while (1) 
	{
		k_msleep(1000);
		    // Set the current date and time
    	rtc_get_time_date(&getcurrent_time);
		print_time_date(&getcurrent_time);

        // k_sem_take(&RTC_sem, K_FOREVER);
        // if(occured == true)
        // {
        //     printf("occured.\n");

        //     clear_rtc_alarm_interrupt();
        // }

		// check_alarm_status();
        // Optionally, you might want to add a delay or other processing here
    }

	return 0;
}
